
######################################
#    Lattice parameters              #
######################################

# Arbitrary conventions used for std_calibration and skew_calibration

NAN_DIVIDEND = 2000  # Used to assign odds to longshots where there is no bid

ALT_L     = 500
ALT_UNIT  = 0.1
ALT_SCALE = 1.0
ALT_A     = 0.5   # Previously 1.0

STD_L        = 500
STD_UNIT     = 0.1
STD_SCALE    = 1.0
STD_A        = 0.0